#ifndef MARK_H_62B23520_7C8E_11DE_8A39_0800200C9A66
#define MARK_H_62B23520_7C8E_11DE_8A39_0800200C9A66

#if defined(_MSC_VER) || (defined(__GNUC__) && (__GNUC__ == 3 && __GNUC_MINOR__ >= 4) || (__GNUC__ >= 4)) // GCC supports "pragma once" correctly since 3.4
#pragma once
#endif


#include "yaml-cpp/dll.h"

namespace YAML
{
	struct YAML_CPP_API Mark {
		Mark(): pos(0), line(0), column(0) {}
		
		static const Mark null_mark() { return Mark(-1, -1, -1); }
		
		int pos;
		int line, column;
		
	private:
		Mark(int pos_, int line_, int column_): pos(pos_), line(line_), column(column_) {}
	};
}

#endif // MARK_H_62B23520_7C8E_11DE_8A39_0800200C9A66
